import importlib
import logging
from typing import Any, Optional, Type

from fastapi import FastAPI
from opentelemetry import _logs, metrics, trace
from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter,
)
from opentelemetry.instrumentation.fastapi import (  # type: ignore
    FastAPIInstrumentor,
)
from opentelemetry.metrics import NoOpMeterProvider
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import NoOpTracerProvider
from typing_extensions import Dict

from beamlit.authentication import get_authentication_headers

from .settings import get_settings

tracer: trace.Tracer | None = None
meter: metrics.Meter | None = None
logger: LoggerProvider | None = None

log = logging.getLogger(__name__)


def auth_headers() -> Dict[str, str]:
    settings = get_settings()
    headers = get_authentication_headers(settings)
    return {
        "x-beamlit-authorization": headers.get("X-Beamlit-Authorization", ""),
        "x-beamlit-workspace": headers.get("X-Beamlit-Workspace", ""),
    }


def get_logger() -> LoggerProvider:
    if logger is None:
        raise Exception("Logger is not initialized")
    return logger


def get_resource_attributes() -> Dict[str, Any]:
    resources = Resource.create()
    resources_dict: Dict[str, Any] = {}
    for key in resources.attributes:
        resources_dict[key] = resources.attributes[key]
    settings = get_settings()
    resources_dict["workspace"] = settings.workspace
    resources_dict["service.name"] = settings.name
    return resources_dict


def get_metrics_exporter() -> OTLPMetricExporter | None:
    settings = get_settings()
    if not settings.enable_opentelemetry:
        return None
    return OTLPMetricExporter(headers=auth_headers())


def get_span_exporter() -> OTLPSpanExporter | None:
    settings = get_settings()
    if not settings.enable_opentelemetry:
        return None
    return OTLPSpanExporter(headers=auth_headers())


def get_log_exporter() -> OTLPLogExporter | None:
    settings = get_settings()
    if not settings.enable_opentelemetry:
        return None
    return OTLPLogExporter(headers=auth_headers())


def _import_class(module_path: str, class_name: str) -> Optional[Type]:  # type: ignore
    """Dynamically import a class from a module path."""
    try:
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        log.error(f"Could not import {class_name} from {module_path}: {str(e)}")
        return None


# Define mapping of instrumentor info: (module path, class name, required package)
INSTRUMENTOR_CONFIGS = {
    "httpx": (
        "opentelemetry.instrumentation.httpx",
        "HTTPXClientInstrumentor",
        "httpx",
    ),
    "anthropic": (
        "opentelemetry.instrumentation.anthropic",
        "AnthropicInstrumentor",
        "anthropic",
    ),
    "chroma": (
        "opentelemetry.instrumentation.chroma",
        "ChromaInstrumentor",
        "chromadb",
    ),
    "cohere": (
        "opentelemetry.instrumentation.cohere",
        "CohereInstrumentor",
        "cohere",
    ),
    "groq": ("opentelemetry.instrumentation.groq", "GroqInstrumentor", "groq"),
    "lance": (
        "opentelemetry.instrumentation.lance",
        "LanceInstrumentor",
        "pylance",
    ),
    "langchain": (
        "opentelemetry.instrumentation.langchain",
        "LangchainInstrumentor",
        "langchain",
    ),
    "llama_index": (
        "opentelemetry.instrumentation.llama_index",
        "LlamaIndexInstrumentor",
        "llama_index",
    ),
    "marqo": (
        "opentelemetry.instrumentation.marqo",
        "MarqoInstrumentor",
        "marqo",
    ),
    "milvus": (
        "opentelemetry.instrumentation.milvus",
        "MilvusInstrumentor",
        "pymilvus",
    ),
    "mistralai": (
        "opentelemetry.instrumentation.mistralai",
        "MistralAiInstrumentor",
        "mistralai",
    ),
    "ollama": (
        "opentelemetry.instrumentation.ollama",
        "OllamaInstrumentor",
        "ollama",
    ),
    "openai": (
        "opentelemetry.instrumentation.openai",
        "OpenAIInstrumentor",
        "openai",
    ),
    "pinecone": (
        "opentelemetry.instrumentation.pinecone",
        "PineconeInstrumentor",
        "pinecone",
    ),
    "qdrant": (
        "opentelemetry.instrumentation.qdrant",
        "QdrantInstrumentor",
        "qdrant_client",
    ),
    "replicate": (
        "opentelemetry.instrumentation.replicate",
        "ReplicateInstrumentor",
        "replicate",
    ),
    "together": (
        "opentelemetry.instrumentation.together",
        "TogetherAiInstrumentor",
        "together",
    ),
    "watsonx": (
        "opentelemetry.instrumentation.watsonx",
        "WatsonxInstrumentor",
        "ibm_watson_machine_learning",
    ),
    "weaviate": (
        "opentelemetry.instrumentation.weaviate",
        "WeaviateInstrumentor",
        "weaviate",
    ),
}


def _is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    try:
        importlib.import_module(package_name)
        return True
    except (ImportError, ModuleNotFoundError):
        return False


def instrument_app(app: FastAPI):
    global tracer
    global meter
    settings = get_settings()
    if not settings.enable_opentelemetry:
        # Use NoOp implementations to stub tracing and metrics
        trace.set_tracer_provider(NoOpTracerProvider())
        tracer = trace.get_tracer(__name__)

        metrics.set_meter_provider(NoOpMeterProvider())
        meter = metrics.get_meter(__name__)
        return

    resource = Resource.create(
        {
            "service.name": settings.name,
            "service.namespace": settings.workspace,
            "service.workspace": settings.workspace,
        }
    )

    # Set up the TracerProvider if not already set
    if not isinstance(trace.get_tracer_provider(), TracerProvider):
        trace_provider = TracerProvider(resource=resource)
        span_processor = BatchSpanProcessor(get_span_exporter())  # type: ignore
        trace_provider.add_span_processor(span_processor)
        trace.set_tracer_provider(trace_provider)
        tracer = trace_provider.get_tracer(__name__)
    else:
        tracer = trace.get_tracer(__name__)

    # Set up the MeterProvider if not already set
    if not isinstance(metrics.get_meter_provider(), MeterProvider):
        metrics_exporter = PeriodicExportingMetricReader(get_metrics_exporter())  # type: ignore
        meter_provider = MeterProvider(
            resource=resource, metric_readers=[metrics_exporter]
        )
        metrics.set_meter_provider(meter_provider)
        meter = meter_provider.get_meter(__name__)
    else:
        meter = metrics.get_meter(__name__)

    if not isinstance(_logs.get_logger_provider(), LoggerProvider):
        logger_provider = LoggerProvider(resource=resource)
        set_logger_provider(logger_provider)
        logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(get_log_exporter())  # type: ignore
        )
        handler = LoggingHandler(
            level=logging.NOTSET, logger_provider=logger_provider
        )
        logging.getLogger().addHandler(handler)
    else:
        logger_provider = _logs.get_logger_provider()

    # Only instrument the app when OpenTelemetry is enabled
    FastAPIInstrumentor.instrument_app(app)  # type: ignore

    for name, (
        module_path,
        class_name,
        required_package,
    ) in INSTRUMENTOR_CONFIGS.items():
        if _is_package_installed(required_package):
            instrumentor_class = _import_class(module_path, class_name)  # type: ignore
            if instrumentor_class:
                try:
                    instrumentor_class().instrument()
                    log.info(f"Successfully instrumented {name}")
                except Exception as e:
                    log.error(f"Failed to instrument {name}: {str(e)}")
            else:
                log.error(f"Could not load instrumentor for {name}")
        else:
            log.debug(
                f"Skipping {name} instrumentation - required package '{required_package}' not installed"
            )


def shutdown_instrumentation():
    if tracer is not None:
        trace_provider = trace.get_tracer_provider()
        if isinstance(trace_provider, TracerProvider):
            trace_provider.shutdown()
    if meter is not None:
        meter_provider = metrics.get_meter_provider()
        if isinstance(meter_provider, MeterProvider):
            meter_provider.shutdown()



import os
from typing import Tuple, Type, Union

from langchain_core.language_models.chat_models import BaseChatModel
from langgraph.graph.graph import CompiledGraph
from pydantic import Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)

from beamlit.common.logger import init as init_logger
from beamlit.models import Agent, Function, Model

global SETTINGS
SETTINGS = None

class SettingsAgent(BaseSettings):
    agent: Union[None, CompiledGraph] = None
    chain: Union[None, list[Agent]] = None
    model: Union[None, Model] = None
    functions: Union[None, list[Function]] = None
    functions_directory: str = Field(default="src/functions")
    chat_model: Union[None, BaseChatModel] = None
    module: str = Field(default="main.main")


class SettingsAuthenticationClient(BaseSettings):
    credentials: Union[None, str] = None


class SettingsAuthentication(BaseSettings):
    apiKey: Union[None, str] = None
    jwt: Union[None, str] = None
    client: SettingsAuthenticationClient = SettingsAuthenticationClient()


class SettingsServer(BaseSettings):
    module: str = Field(default="main.main")
    port: int = Field(default=80)
    host: str = Field(default="0.0.0.0")
    directory: str = Field(default="src")

class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        yaml_file="beamlit.yaml",
        env_prefix="bl_",
        env_nested_delimiter="_",
        extra="ignore",
    )

    workspace: str
    environment: str = Field(default="production")
    remote: bool = Field(default=False)
    type: str = Field(default="agent")
    name: str = Field(default="beamlit-agent")
    base_url: str = Field(default="https://api.beamlit.com/v0")
    app_url: str = Field(default="https://app.beamlit.com")
    run_url: str = Field(default="https://run.beamlit.com")
    mcp_hub_url: str = Field(default="https://mcp-hub-server.beamlit.workers.com")
    registry_url: str = Field(default="https://us.registry.beamlit.com")
    log_level: str = Field(default="INFO")
    enable_opentelemetry: bool = Field(default=False)
    agent: SettingsAgent = SettingsAgent()
    server: SettingsServer = SettingsServer()
    authentication: SettingsAuthentication = SettingsAuthentication()

    def __init__(self, **data):
        super().__init__(**data)
        if os.getenv('BL_ENV') == 'dev':
            self.base_url = os.getenv('BL_BASE_URL') or "https://api.beamlit.dev/v0"
            self.run_url = os.getenv('BL_RUN_URL') or "https://run.beamlit.dev"
            self.mcp_hub_url = os.getenv('BL_MCP_HUB_URL') or "https://mcp-hub-server.beamlit.workers.dev"
            self.registry_url = os.getenv('BL_REGISTRY_URL') or "https://eu.registry.beamlit.dev"
            self.app_url = os.getenv('BL_APP_URL') or "https://app.beamlit.dev"
            
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (
            env_settings,
            dotenv_settings,
            file_secret_settings,
            YamlConfigSettingsSource(settings_cls),
            init_settings,
        )

def get_settings() -> Settings:
    return SETTINGS

def init() -> Settings:
    """Parse the beamlit.yaml file to get configurations."""
    from beamlit.authentication.credentials import current_context

    global SETTINGS

    context = current_context()
    kwargs = {}
    if context.workspace:
        kwargs["workspace"] = context.workspace
    if context.environment:
        kwargs["environment"] = context.environment

    SETTINGS = Settings(**kwargs)
    init_logger(SETTINGS.log_level)

    return SETTINGS



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="QPS")


@_attrs_define
class QPS:
    """Query per second per element, can be per response status code (e.g. 200, 400) or per location

    Attributes:
        region_code (Union[Unset, float]): QPS for location
    """

    region_code: Union[Unset, float] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        region_code = self.region_code

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if region_code is not UNSET:
            field_dict["region_code"] = region_code

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        region_code = d.pop("region_code", UNSET)

        qps = cls(
            region_code=region_code,
        )

        qps.additional_properties = d
        return qps

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.integration_connection_config import IntegrationConnectionConfig
    from ..models.integration_connection_secret import IntegrationConnectionSecret


T = TypeVar("T", bound="IntegrationConnectionSpec")


@_attrs_define
class IntegrationConnectionSpec:
    """Integration connection specification

    Attributes:
        config (Union[Unset, IntegrationConnectionConfig]): Integration config
        integration (Union[Unset, str]): Integration type
        sandbox (Union[Unset, bool]): Sandbox mode
        secret (Union[Unset, IntegrationConnectionSecret]): Integration secret
    """

    config: Union[Unset, "IntegrationConnectionConfig"] = UNSET
    integration: Union[Unset, str] = UNSET
    sandbox: Union[Unset, bool] = UNSET
    secret: Union[Unset, "IntegrationConnectionSecret"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        config: Union[Unset, dict[str, Any]] = UNSET
        if self.config and not isinstance(self.config, Unset):
            config = self.config.to_dict()

        integration = self.integration

        sandbox = self.sandbox

        secret: Union[Unset, dict[str, Any]] = UNSET
        if self.secret and not isinstance(self.secret, Unset):
            secret = self.secret.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if config is not UNSET:
            field_dict["config"] = config
        if integration is not UNSET:
            field_dict["integration"] = integration
        if sandbox is not UNSET:
            field_dict["sandbox"] = sandbox
        if secret is not UNSET:
            field_dict["secret"] = secret

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.integration_connection_config import IntegrationConnectionConfig
        from ..models.integration_connection_secret import IntegrationConnectionSecret

        if not src_dict:
            return None
        d = src_dict.copy()
        _config = d.pop("config", UNSET)
        config: Union[Unset, IntegrationConnectionConfig]
        if isinstance(_config, Unset):
            config = UNSET
        else:
            config = IntegrationConnectionConfig.from_dict(_config)

        integration = d.pop("integration", UNSET)

        sandbox = d.pop("sandbox", UNSET)

        _secret = d.pop("secret", UNSET)
        secret: Union[Unset, IntegrationConnectionSecret]
        if isinstance(_secret, Unset):
            secret = UNSET
        else:
            secret = IntegrationConnectionSecret.from_dict(_secret)

        integration_connection_spec = cls(
            config=config,
            integration=integration,
            sandbox=sandbox,
            secret=secret,
        )

        integration_connection_spec.additional_properties = d
        return integration_connection_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import base64
import json
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Dict, Generator, Optional

from httpx import Auth, Request, Response, post


@dataclass
class DeviceLogin:
    client_id: str
    scope: str


@dataclass
class DeviceLoginResponse:
    client_id: str
    device_code: str
    user_code: str
    expires_in: int
    interval: int
    verification_uri: str
    verification_uri_complete: str


@dataclass
class DeviceLoginFinalizeRequest:
    grant_type: str
    client_id: str
    device_code: str


@dataclass
class DeviceLoginFinalizeResponse:
    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str


class BearerToken(Auth):
    def __init__(self, credentials, workspace_name: str, base_url: str):
        self.credentials = credentials
        self.workspace_name = workspace_name
        self.base_url = base_url

    def get_headers(self) -> Dict[str, str]:
        err = self.refresh_if_needed()
        if err:
            raise err
        return {
            "X-Beamlit-Authorization": f"Bearer {self.credentials.access_token}",
            "X-Beamlit-Workspace": self.workspace_name,
        }

    def refresh_if_needed(self) -> Optional[Exception]:
        # Need to refresh token if expires in less than 10 minutes
        parts = self.credentials.access_token.split(".")
        if len(parts) != 3:
            return Exception("Invalid JWT token format")

        try:
            claims_bytes = base64.urlsafe_b64decode(parts[1] + "=" * (-len(parts[1]) % 4))
            claims = json.loads(claims_bytes)
        except Exception as e:
            return Exception(f"Failed to decode/parse JWT claims: {str(e)}")
        exp_time = datetime.fromtimestamp(claims["exp"])
        current_time = datetime.now()
        # Refresh if token expires in less than 10 minutes
        if current_time + timedelta(minutes=10) > exp_time:
            return self.do_refresh()

        return None

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        err = self.refresh_if_needed()
        if err:
            return err

        request.headers["X-Beamlit-Authorization"] = f"Bearer {self.credentials.access_token}"
        request.headers["X-Beamlit-Workspace"] = self.workspace_name
        yield request

    def do_refresh(self) -> Optional[Exception]:
        if not self.credentials.refresh_token:
            return Exception("No refresh token to refresh")

        url = f"{self.base_url}/oauth/token"
        refresh_data = {
            "grant_type": "refresh_token",
            "refresh_token": self.credentials.refresh_token,
            "device_code": self.credentials.device_code,
            "client_id": "beamlit",
        }

        try:
            response = post(url, json=refresh_data, headers={"Content-Type": "application/json"})
            response.raise_for_status()
            finalize_response = DeviceLoginFinalizeResponse(**response.json())

            if not finalize_response.refresh_token:
                finalize_response.refresh_token = self.credentials.refresh_token

            from .credentials import Credentials, save_credentials

            creds = Credentials(
                access_token=finalize_response.access_token,
                refresh_token=finalize_response.refresh_token,
                expires_in=finalize_response.expires_in,
                device_code=self.credentials.device_code,
            )

            self.credentials = creds
            save_credentials(self.workspace_name, creds)
            return None

        except Exception as e:
            return Exception(f"Failed to refresh token: {str(e)}")



"""Contains all the data models used in inputs/outputs"""

from .acl import ACL
from .agent import Agent
from .agent_chain import AgentChain
from .agent_history import AgentHistory
from .agent_history_event import AgentHistoryEvent
from .agent_metadata import AgentMetadata
from .agent_release import AgentRelease
from .agent_spec import AgentSpec
from .api_key import ApiKey
from .configuration import Configuration
from .continent import Continent
from .core_event import CoreEvent
from .core_spec import CoreSpec
from .core_spec_configurations import CoreSpecConfigurations
from .country import Country
from .create_api_key_for_service_account_body import CreateApiKeyForServiceAccountBody
from .create_workspace_service_account_body import CreateWorkspaceServiceAccountBody
from .create_workspace_service_account_response_200 import CreateWorkspaceServiceAccountResponse200
from .delete_workspace_service_account_response_200 import DeleteWorkspaceServiceAccountResponse200
from .environment import Environment
from .environment_metadata import EnvironmentMetadata
from .environment_metrics import EnvironmentMetrics
from .environment_spec import EnvironmentSpec
from .flavor import Flavor
from .function import Function
from .function_kit import FunctionKit
from .function_metadata import FunctionMetadata
from .function_release import FunctionRelease
from .function_spec import FunctionSpec
from .get_trace_ids_response_200 import GetTraceIdsResponse200
from .get_trace_logs_response_200 import GetTraceLogsResponse200
from .get_trace_response_200 import GetTraceResponse200
from .get_workspace_service_accounts_response_200_item import (
    GetWorkspaceServiceAccountsResponse200Item,
)
from .histogram_bucket import HistogramBucket
from .histogram_stats import HistogramStats
from .increase_and_rate_metric import IncreaseAndRateMetric
from .integration_config import IntegrationConfig
from .integration_connection import IntegrationConnection
from .integration_connection_config import IntegrationConnectionConfig
from .integration_connection_secret import IntegrationConnectionSecret
from .integration_connection_spec import IntegrationConnectionSpec
from .integration_model import IntegrationModel
from .integration_repository import IntegrationRepository
from .invite_workspace_user_body import InviteWorkspaceUserBody
from .last_n_requests_metric import LastNRequestsMetric
from .latency_metric import LatencyMetric
from .location_response import LocationResponse
from .metadata import Metadata
from .metadata_labels import MetadataLabels
from .metric import Metric
from .metrics import Metrics
from .metrics_models import MetricsModels
from .metrics_request_total_per_code import MetricsRequestTotalPerCode
from .metrics_rps_per_code import MetricsRpsPerCode
from .model import Model
from .model_metadata import ModelMetadata
from .model_private_cluster import ModelPrivateCluster
from .model_release import ModelRelease
from .model_spec import ModelSpec
from .owner_fields import OwnerFields
from .pending_invitation import PendingInvitation
from .pending_invitation_accept import PendingInvitationAccept
from .pending_invitation_render import PendingInvitationRender
from .pending_invitation_render_invited_by import PendingInvitationRenderInvitedBy
from .pending_invitation_render_workspace import PendingInvitationRenderWorkspace
from .pending_invitation_workspace_details import PendingInvitationWorkspaceDetails
from .pod_template_spec import PodTemplateSpec
from .policy import Policy
from .policy_location import PolicyLocation
from .policy_spec import PolicySpec
from .private_cluster import PrivateCluster
from .private_location import PrivateLocation
from .provider_config import ProviderConfig
from .qps import QPS
from .repository import Repository
from .request_duration_over_time_metric import RequestDurationOverTimeMetric
from .request_duration_over_time_metrics import RequestDurationOverTimeMetrics
from .request_total_by_origin_metric import RequestTotalByOriginMetric
from .request_total_by_origin_metric_request_total_by_origin import (
    RequestTotalByOriginMetricRequestTotalByOrigin,
)
from .request_total_by_origin_metric_request_total_by_origin_and_code import (
    RequestTotalByOriginMetricRequestTotalByOriginAndCode,
)
from .request_total_metric import RequestTotalMetric
from .request_total_metric_request_total_per_code import RequestTotalMetricRequestTotalPerCode
from .request_total_metric_rps_per_code import RequestTotalMetricRpsPerCode
from .resource_deployment_metrics import ResourceDeploymentMetrics
from .resource_deployment_metrics_inference_per_second_per_region import (
    ResourceDeploymentMetricsInferencePerSecondPerRegion,
)
from .resource_deployment_metrics_query_per_second_per_region_per_code import (
    ResourceDeploymentMetricsQueryPerSecondPerRegionPerCode,
)
from .resource_environment_metrics import ResourceEnvironmentMetrics
from .resource_environment_metrics_request_total_per_code import (
    ResourceEnvironmentMetricsRequestTotalPerCode,
)
from .resource_environment_metrics_rps_per_code import ResourceEnvironmentMetricsRpsPerCode
from .resource_log import ResourceLog
from .resource_metrics import ResourceMetrics
from .runtime import Runtime
from .runtime_readiness_probe import RuntimeReadinessProbe
from .runtime_resources import RuntimeResources
from .serverless_config import ServerlessConfig
from .spec_configuration import SpecConfiguration
from .store_agent import StoreAgent
from .store_agent_labels import StoreAgentLabels
from .store_configuration import StoreConfiguration
from .store_configuration_option import StoreConfigurationOption
from .store_function import StoreFunction
from .store_function_kit import StoreFunctionKit
from .store_function_labels import StoreFunctionLabels
from .store_function_parameter import StoreFunctionParameter
from .time_fields import TimeFields
from .token_rate_metric import TokenRateMetric
from .token_rate_metrics import TokenRateMetrics
from .token_total_metric import TokenTotalMetric
from .trace_ids_response import TraceIdsResponse
from .update_workspace_service_account_body import UpdateWorkspaceServiceAccountBody
from .update_workspace_service_account_response_200 import UpdateWorkspaceServiceAccountResponse200
from .update_workspace_user_role_body import UpdateWorkspaceUserRoleBody
from .websocket_channel import WebsocketChannel
from .workspace import Workspace
from .workspace_labels import WorkspaceLabels
from .workspace_user import WorkspaceUser

__all__ = (
    "ACL",
    "Agent",
    "AgentChain",
    "AgentHistory",
    "AgentHistoryEvent",
    "AgentMetadata",
    "AgentRelease",
    "AgentSpec",
    "ApiKey",
    "Configuration",
    "Continent",
    "CoreEvent",
    "CoreSpec",
    "CoreSpecConfigurations",
    "Country",
    "CreateApiKeyForServiceAccountBody",
    "CreateWorkspaceServiceAccountBody",
    "CreateWorkspaceServiceAccountResponse200",
    "DeleteWorkspaceServiceAccountResponse200",
    "Environment",
    "EnvironmentMetadata",
    "EnvironmentMetrics",
    "EnvironmentSpec",
    "Flavor",
    "Function",
    "FunctionKit",
    "FunctionMetadata",
    "FunctionRelease",
    "FunctionSpec",
    "GetTraceIdsResponse200",
    "GetTraceLogsResponse200",
    "GetTraceResponse200",
    "GetWorkspaceServiceAccountsResponse200Item",
    "HistogramBucket",
    "HistogramStats",
    "IncreaseAndRateMetric",
    "IntegrationConfig",
    "IntegrationConnection",
    "IntegrationConnectionConfig",
    "IntegrationConnectionSecret",
    "IntegrationConnectionSpec",
    "IntegrationModel",
    "IntegrationRepository",
    "InviteWorkspaceUserBody",
    "LastNRequestsMetric",
    "LatencyMetric",
    "LocationResponse",
    "Metadata",
    "MetadataLabels",
    "Metric",
    "Metrics",
    "MetricsModels",
    "MetricsRequestTotalPerCode",
    "MetricsRpsPerCode",
    "Model",
    "ModelMetadata",
    "ModelPrivateCluster",
    "ModelRelease",
    "ModelSpec",
    "OwnerFields",
    "PendingInvitation",
    "PendingInvitationAccept",
    "PendingInvitationRender",
    "PendingInvitationRenderInvitedBy",
    "PendingInvitationRenderWorkspace",
    "PendingInvitationWorkspaceDetails",
    "PodTemplateSpec",
    "Policy",
    "PolicyLocation",
    "PolicySpec",
    "PrivateCluster",
    "PrivateLocation",
    "ProviderConfig",
    "QPS",
    "Repository",
    "RequestDurationOverTimeMetric",
    "RequestDurationOverTimeMetrics",
    "RequestTotalByOriginMetric",
    "RequestTotalByOriginMetricRequestTotalByOrigin",
    "RequestTotalByOriginMetricRequestTotalByOriginAndCode",
    "RequestTotalMetric",
    "RequestTotalMetricRequestTotalPerCode",
    "RequestTotalMetricRpsPerCode",
    "ResourceDeploymentMetrics",
    "ResourceDeploymentMetricsInferencePerSecondPerRegion",
    "ResourceDeploymentMetricsQueryPerSecondPerRegionPerCode",
    "ResourceEnvironmentMetrics",
    "ResourceEnvironmentMetricsRequestTotalPerCode",
    "ResourceEnvironmentMetricsRpsPerCode",
    "ResourceLog",
    "ResourceMetrics",
    "Runtime",
    "RuntimeReadinessProbe",
    "RuntimeResources",
    "ServerlessConfig",
    "SpecConfiguration",
    "StoreAgent",
    "StoreAgentLabels",
    "StoreConfiguration",
    "StoreConfigurationOption",
    "StoreFunction",
    "StoreFunctionKit",
    "StoreFunctionLabels",
    "StoreFunctionParameter",
    "TimeFields",
    "TokenRateMetric",
    "TokenRateMetrics",
    "TokenTotalMetric",
    "TraceIdsResponse",
    "UpdateWorkspaceServiceAccountBody",
    "UpdateWorkspaceServiceAccountResponse200",
    "UpdateWorkspaceUserRoleBody",
    "WebsocketChannel",
    "Workspace",
    "WorkspaceLabels",
    "WorkspaceUser",
)



import asyncio
import warnings
from dataclasses import dataclass
from typing import Callable

import pydantic
import typing_extensions as t
from langchain_core.tools.base import BaseTool, ToolException

from beamlit.api.functions import get_function
from beamlit.authentication.authentication import AuthenticatedClient
from beamlit.common.settings import get_settings
from beamlit.models import Function, StoreFunctionParameter
from beamlit.run import RunClient


def create_dynamic_schema(name: str, parameters: list[StoreFunctionParameter]) -> type[pydantic.BaseModel]:
    field_definitions = {}
    for param in parameters:
        field_type = str
        if param.type_ == "number":
            field_type = float
        elif param.type_ == "integer":
            field_type = int
        elif param.type_ == "boolean":
            field_type = bool

        field_definitions[param.name] = (
            field_type,
            pydantic.Field(description=param.description or "")
        )
    return pydantic.create_model(
        f"{name}Schema",
        **field_definitions
    )


class RemoteTool(BaseTool):
    """
    Remote tool
    """

    client: RunClient
    resource_name: str
    kit: bool = False
    handle_tool_error: bool | str | Callable[[ToolException], str] | None = True

    @t.override
    def _run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        warnings.warn(
            "Invoke this tool asynchronousely using `ainvoke`. This method exists only to satisfy standard tests.",
            stacklevel=1,
        )
        return asyncio.run(self._arun(*args, **kwargs))

    @t.override
    async def _arun(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        settings = get_settings()
        body = {**kwargs}
        if self.kit:
            body["name"] = self.name
        result = self.client.run(
            "function",
            self.resource_name,
            settings.environment,
            "POST",
            json=body
        )
        return result.text

    @t.override
    @property
    def tool_call_schema(self) -> type[pydantic.BaseModel]:
        assert self.args_schema is not None  # noqa: S101
        return self.args_schema

@dataclass
class RemoteToolkit:
    """
    Remote toolkit
    """

    client: AuthenticatedClient
    function: str
    _function: Function | None = None

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def initialize(self) -> None:
        """Initialize the session and retrieve tools list"""
        if self._function is None:
            self._function = get_function.sync_detailed(self.function, client=self.client).parsed

    @t.override
    def get_tools(self) -> list[BaseTool]:
        if self._function is None:
            raise RuntimeError("Must initialize the toolkit first")

        if self._function.spec.kit:
            return [
                RemoteTool(
                    client=RunClient(self.client),
                    name=func.name,
                    resource_name=self._function.metadata.name,
                    kit=True,
                    description=func.description or "",
                    args_schema=create_dynamic_schema(func.name, func.parameters),
                )
                for func in self._function.spec.kit
            ]

        return [
            RemoteTool(
                client=RunClient(self.client),
                name=self._function.metadata.name,
                resource_name=self._function.metadata.name,
                description=self._function.spec.description or "",
                args_schema=create_dynamic_schema(
                    self._function.metadata.name,
                    self._function.spec.parameters
                ),
            )
        ]


from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.agent import Agent
from ...types import UNSET, Response


def _get_kwargs(
    agent_name: str,
    *,
    environment: str,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["environment"] = environment

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "delete",
        "url": f"/agents/{agent_name}",
        "params": params,
    }

    return _kwargs


def _parse_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> Optional[Agent]:
    if response.status_code == 200:
        response_200 = Agent.from_dict(response.json())

        return response_200
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> Response[Agent]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    agent_name: str,
    *,
    client: AuthenticatedClient,
    environment: str,
) -> Response[Agent]:
    """Delete agent by name

    Args:
        agent_name (str):
        environment (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Agent]
    """

    kwargs = _get_kwargs(
        agent_name=agent_name,
        environment=environment,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    agent_name: str,
    *,
    client: AuthenticatedClient,
    environment: str,
) -> Optional[Agent]:
    """Delete agent by name

    Args:
        agent_name (str):
        environment (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Agent
    """

    return sync_detailed(
        agent_name=agent_name,
        client=client,
        environment=environment,
    ).parsed


async def asyncio_detailed(
    agent_name: str,
    *,
    client: AuthenticatedClient,
    environment: str,
) -> Response[Agent]:
    """Delete agent by name

    Args:
        agent_name (str):
        environment (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Agent]
    """

    kwargs = _get_kwargs(
        agent_name=agent_name,
        environment=environment,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    agent_name: str,
    *,
    client: AuthenticatedClient,
    environment: str,
) -> Optional[Agent]:
    """Delete agent by name

    Args:
        agent_name (str):
        environment (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Agent
    """

    return (
        await asyncio_detailed(
            agent_name=agent_name,
            client=client,
            environment=environment,
        )
    ).parsed



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.histogram_bucket import HistogramBucket
    from ..models.histogram_stats import HistogramStats


T = TypeVar("T", bound="LatencyMetric")


@_attrs_define
class LatencyMetric:
    """Latency metrics

    Attributes:
        global_histogram (Union[Unset, HistogramBucket]): Histogram bucket
        global_stats (Union[Unset, HistogramStats]): Histogram stats
        histogram_per_code (Union[Unset, HistogramBucket]): Histogram bucket
        stats_per_code (Union[Unset, HistogramStats]): Histogram stats
    """

    global_histogram: Union[Unset, "HistogramBucket"] = UNSET
    global_stats: Union[Unset, "HistogramStats"] = UNSET
    histogram_per_code: Union[Unset, "HistogramBucket"] = UNSET
    stats_per_code: Union[Unset, "HistogramStats"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        global_histogram: Union[Unset, dict[str, Any]] = UNSET
        if self.global_histogram and not isinstance(self.global_histogram, Unset):
            global_histogram = self.global_histogram.to_dict()

        global_stats: Union[Unset, dict[str, Any]] = UNSET
        if self.global_stats and not isinstance(self.global_stats, Unset):
            global_stats = self.global_stats.to_dict()

        histogram_per_code: Union[Unset, dict[str, Any]] = UNSET
        if self.histogram_per_code and not isinstance(self.histogram_per_code, Unset):
            histogram_per_code = self.histogram_per_code.to_dict()

        stats_per_code: Union[Unset, dict[str, Any]] = UNSET
        if self.stats_per_code and not isinstance(self.stats_per_code, Unset):
            stats_per_code = self.stats_per_code.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if global_histogram is not UNSET:
            field_dict["globalHistogram"] = global_histogram
        if global_stats is not UNSET:
            field_dict["globalStats"] = global_stats
        if histogram_per_code is not UNSET:
            field_dict["histogramPerCode"] = histogram_per_code
        if stats_per_code is not UNSET:
            field_dict["statsPerCode"] = stats_per_code

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.histogram_bucket import HistogramBucket
        from ..models.histogram_stats import HistogramStats

        if not src_dict:
            return None
        d = src_dict.copy()
        _global_histogram = d.pop("globalHistogram", UNSET)
        global_histogram: Union[Unset, HistogramBucket]
        if isinstance(_global_histogram, Unset):
            global_histogram = UNSET
        else:
            global_histogram = HistogramBucket.from_dict(_global_histogram)

        _global_stats = d.pop("globalStats", UNSET)
        global_stats: Union[Unset, HistogramStats]
        if isinstance(_global_stats, Unset):
            global_stats = UNSET
        else:
            global_stats = HistogramStats.from_dict(_global_stats)

        _histogram_per_code = d.pop("histogramPerCode", UNSET)
        histogram_per_code: Union[Unset, HistogramBucket]
        if isinstance(_histogram_per_code, Unset):
            histogram_per_code = UNSET
        else:
            histogram_per_code = HistogramBucket.from_dict(_histogram_per_code)

        _stats_per_code = d.pop("statsPerCode", UNSET)
        stats_per_code: Union[Unset, HistogramStats]
        if isinstance(_stats_per_code, Unset):
            stats_per_code = UNSET
        else:
            stats_per_code = HistogramStats.from_dict(_stats_per_code)

        latency_metric = cls(
            global_histogram=global_histogram,
            global_stats=global_stats,
            histogram_per_code=histogram_per_code,
            stats_per_code=stats_per_code,
        )

        latency_metric.additional_properties = d
        return latency_metric

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
